/*
 * Copyright 2019, OpenCensus Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.opencensus.contrib.http.jaxrs;

import javax.ws.rs.container.ContainerRequestContext;
import javax.ws.rs.container.ResourceInfo;

/** Contains both {@link ContainerRequestContext} and {@link ResourceInfo}. */
class ExtendedContainerRequest {
  private final ContainerRequestContext requestContext;
  private final ResourceInfo resourceInfo;

  public ExtendedContainerRequest(
      ContainerRequestContext requestContext, ResourceInfo resourceInfo) {
    this.requestContext = requestContext;
    this.resourceInfo = resourceInfo;
  }

  public ContainerRequestContext getRequestContext() {
    return requestContext;
  }

  public ResourceInfo getResourceInfo() {
    return resourceInfo;
  }
}
