/*
 * Copyright 2017, OpenCensus Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.opencensus.implcore.internal;

/**
 * An {@link EventQueue} that processes events in the current thread. This class can be used for
 * testing.
 */
public class SimpleEventQueue implements EventQueue {

  @Override
  public void enqueue(Entry entry) {
    entry.process();
  }

  @Override
  public void shutdown() {}
}
