/*
 * Copyright 2018, OpenCensus Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.opencensus.contrib.spring.aop;

import java.sql.SQLException;

public class Sample {
  @Traced()
  void example1() {
    // do work
  }

  @Traced(name = "custom-span-name")
  void example2() {
    // do moar work
  }

  @Traced()
  void call(long delay) throws Exception {
    Thread.sleep(delay);
  }

  @Traced(name = "blah")
  void custom(long delay) throws Exception {
    Thread.sleep(delay);
  }

  @Traced()
  void boom() throws Exception {
    throw new Exception("boom");
  }

  public void execute(String sql) throws SQLException {}

  public void executeQuery(String sql) throws SQLException {}

  public void executeUpdate(String sql) throws SQLException {}

  public void executeLargeUpdate(String sql) throws SQLException {}
}
