/*
 * Copyright 2018, OpenCensus Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.opencensus.examples.trace;

import java.util.logging.Logger;

/** Util methods. */
final class Utils {

  private static final Logger logger = Logger.getLogger(Utils.class.getName());

  static void sleep(int ms) {
    // A helper to avoid try-catch when invoking Thread.sleep so that
    // sleeps can be succinct and not permeated by exception handling.
    try {
      Thread.sleep(ms);
    } catch (Exception e) {
      logger.warning((String.format("Failed to sleep for %dms. Exception: %s", ms, e)));
    }
  }

  private Utils() {}
}
