/*
 * Copyright 2018, OpenCensus Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * This package contains the Java implementation of the OpenCensus Agent (OC-Agent) Metrics
 * Exporter.
 *
 * <p>WARNING: Currently all the public classes under this package are marked as {@link
 * io.opencensus.common.ExperimentalApi}. The classes and APIs under {@link
 * io.opencensus.exporter.metrics.ocagent} are likely to get backwards-incompatible updates in the
 * future. DO NOT USE except for experimental purposes.
 *
 * <p>See more details on
 * https://github.com/census-instrumentation/opencensus-proto/tree/master/src/opencensus/proto/agent.
 */
@io.opencensus.common.ExperimentalApi
package io.opencensus.exporter.metrics.ocagent;
