/*
 * Copyright 2017, OpenCensus Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.opencensus.implcore.stats;

import static com.google.common.base.Preconditions.checkNotNull;

import io.opencensus.stats.StatsRecorder;

/** Implementation of {@link StatsRecorder}. */
public final class StatsRecorderImpl extends StatsRecorder {
  private final StatsManager statsManager;

  StatsRecorderImpl(StatsManager statsManager) {
    checkNotNull(statsManager, "StatsManager");
    this.statsManager = statsManager;
  }

  @Override
  public MeasureMapImpl newMeasureMap() {
    return MeasureMapImpl.create(statsManager);
  }
}
